from vertexai.preview import extensions
from .init import init_vertex
from ..logging import log
from ..utils.gcp_project import get_gcp_project
from ..utils.parsers import validate_extension_id
import base64
import json
from io import StringIO

class VertexAIExtensions:
    def __init__(self):
        self.CODE_INTERPRETER_WRITTEN_FILES = []
        self.css_styles = """
        <style>
        .main_summary {
          font-weight: bold;
          font-size: 14px; color: #4285F4;
          background-color:rgba(221, 221, 221, 0.5); padding:8px;}
        </style>
        """
        self.IMAGE_FILE_EXTENSIONS = set(["jpg", "jpeg", "png"])
        self.location = "us-central1"

    def get_extension_import_config(self, display_name: str, description: str,
                                    api_spec_gcs: dict, service_account_name: dict, tool_use_examples: list):
        tool_use_examples = [
            {
                "extensionOperation": {
                    "operationId": "say_hello",
                },
                "displayName": "Say hello in the requested language",
                "query": "Say hello in French",
                "requestParams": {
                    "fields": [
                        {
                            "key": "apiServicePrompt",
                            "value": {
                                "string_value": "French",
                            }
                        }
                    ]
                },
                "responseParams": {
                    "fields": [
                        {
                            "key": "apiServiceOutput",
                            "value": {
                                "string_value": "bonjour",
                            },
                        }
                    ],
                },
                "responseSummary": "Bonjour"
            }
        ]

        return {
            "displayName": display_name,
            "description": description,
            "manifest": {
                "name": "EXTENSION_NAME_LLM",
                "description": "DESCRIPTION_LLM",
                "apiSpec": {
                    "openApiGcsUri": api_spec_gcs,
                },
                "authConfig": {
                    "authType": "OAUTH",
                    "oauthConfig": {"service_account": service_account_name}
                }
            },
            "toolUseExamples": tool_use_examples,
        }

    def create_extension_instance(self, display_name: str, description: str, open_api_gcs_uri: str,
                                  llm_name: str = None, llm_description: str = None, runtime_config: dict = None, service_account: str = None):
        project_id = get_gcp_project()
        extension_name = f"projects/{project_id}/locations/us-central1/extensions/{validate_extension_id(display_name)}"

        extension = extensions.Extension.create(
            extension_name=extension_name,
            display_name=display_name,
            description=description,
            runtime_config=runtime_config or None,
            manifest={
                "name": llm_name or display_name,
                "description": llm_description or description,
                "api_spec": {
                    "open_api_gcs_uri": open_api_gcs_uri
                },
                "auth_config": {
                    "auth_type": "GOOGLE_SERVICE_ACCOUNT_AUTH",
                    "google_service_account_config": service_account or {},
                },
            },
        )
        log.info(f"Created Vertex Extension: {extension_name}")

        return extension

    def execute_extension(self, operation_id: str, operation_params: dict, extension_id: str):
        init_vertex(location=self.location)

        if not extension_id.startswith("projects/"):
            project_id = get_gcp_project()
            extension_name = f"projects/{project_id}/locations/{self.location}/extensions/{extension_id}"
        else:
            extension_name = extension_id

        extension = extensions.Extension(extension_name)

        response = extension.execute(
            operation_id=operation_id,
            operation_params=operation_params,
        )

        return response

    def execute_code_extension(self, query: str, filenames: list[str] = None, gcs_files: list[str] = None):
        if filenames and gcs_files:
            raise ValueError("Can't specify both filenames and gcs_files")

        extension_code_interpreter = extensions.Extension.from_hub("code_interpreter")

        file_arr = None
        if filenames:
            file_arr = [
                {
                    "name": filename,
                    "contents": base64.b64encode(open(filename, "rb").read()).decode()
                }
                for filename in filenames
            ]

        response = extension_code_interpreter.execute(
            operation_id="generate_and_execute",
            operation_params={
                "query": query,
                "files": file_arr,
                "file_gcs_uris": gcs_files
            })

        self.CODE_INTERPRETER_WRITTEN_FILES.extend(
            [item['name'] for item in response['output_files']])

        if response.get('execution_error'):
            log.error(f"Code Execution Response failed with: {response.get('execution_error')} - maybe retry?")

        return response

    def parse_files_to_html(self, outputFiles, save_files_locally=True):
        file_list = []
        details_tml = """<details><summary>{name}</summary><div>{html_content}</div></details>"""

        if not outputFiles:
            return "No Files generated from the code"
        # Sort output_files so images are displayed before other files such as JSON.
        for output_file in sorted(
                outputFiles,
                key=lambda x: x["name"].split(".")[-1] not in self.IMAGE_FILE_EXTENSIONS,
        ):
            file_name = output_file.get("name")
            file_contents = base64.b64decode(output_file.get("contents"))
            if save_files_locally:
                open(file_name, "wb").write(file_contents)

            if file_name.split(".")[-1] in self.IMAGE_FILE_EXTENSIONS:
                # Render Image
                file_html_content = ('<img src="data:image/png;base64, '
                                     f'{output_file.get("contents")}" />')
            elif file_name.endswith(".json"):
                import pprint
                # Pretty print JSON
                json_pp = pprint.pformat(
                    json.loads(file_contents.decode()),
                    compact=False,
                    width=160)
                file_html_content = (f'<span>{json_pp}</span>')
            elif file_name.endswith(".csv"):
                # CSV
                try:
                    import pandas
                except ImportError:
                    log.error("Need pandas for csv processing")
                csv_md = pandas.read_csv(
                    StringIO(file_contents.decode())).to_markdown(index=False)
                file_html_content = f'<span>{csv_md}</span>'
            elif file_name.endswith(".pkl"):
                # PKL
                file_html_content = f'<span>Preview N/A</span>'
            else:
                file_html_content = f"<span>{file_contents.decode()}</span>"

            file_list.append({'name': file_name, "html_content": file_html_content})

        buffer_html = [details_tml.format(**_file) for _file in file_list]
        return "".join(buffer_html)
    
    def process_response(self, response: dict, save_files_locally=None) -> str:
        result_template = """
        <details open>
            <summary class='main_summary'>{summary}:</summary>
            <div><pre>{content}</pre></div>
        </details>
        """

        result = ""
        code = response.get('generated_code')
        if 'execution_result' in response and response['execution_result'] != "":
            result = result_template.format(
                summary="Executed Code Output",
                content=response.get('execution_result'))
        else:
            result = result_template.format(
                summary="Executed Code Output",
                content="Code does not produce printable output.")

        if response.get('execution_error', None):
            result += result_template.format(
                summary="Generated Code Raised a (Possibly Non-Fatal) Exception",
                content=response.get('execution_error', None))

        result += result_template.format(
            summary="Files Created <u>(Click on filename to view content)</u>",
            content=self.parse_files_to_html(
                response.get('output_files', []),
                save_files_locally=True))

        html_content = f"""
{self.css_styles}
<div id='main'>
    <h3>Generated Code by Code Interpreter</h3>
    <pre><code>{code}</code></pre>
    <h3>Code Execution Results</h3>
    {result}
</div>
"""
        if save_files_locally:
            with open('code_execution_results.html', 'w') as file:
                file.write(html_content)

        return html_content


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import pathlib

from ..logging import log
from ..pubsub import process_pubsub_message
from .message_data import handle_gcs_message, handle_google_drive_message, handle_github_message, handle_http_message, handle_json_content_message
from .publish import process_docs_chunks_vector_name
from .splitter import chunk_doc_to_docs

from ..llamaindex.import_files import llamaindex_chunker_check
from ..discovery_engine.chunker_handler import discovery_engine_chunker_check

from . import loaders

def direct_file_to_embed(file_name: pathlib.Path, metadata: dict, vector_name: str):
    """
    Send direct files to chunking embed pipeline

    
    
    """
    log.info(f"Sending direct file upload {file_name} to loaders.read_file_to_documents {metadata}")
    docs = loaders.read_file_to_documents(file_name, metadata=metadata)
    if docs is None:
        log.warning(f"loaders.read_file_to_documents docs2 failed to load file {metadata}")

        return None

    chunks = chunk_doc_to_docs(docs, file_name.suffix, vector_name=vector_name)
    
    return format_chunk_return(chunks, metadata, vector_name)



def data_to_embed_pubsub(data: dict):
    """Triggered from a message on a Cloud Pub/Sub topic.
    Args:
         data JSON
    """

    message_data, metadata, vector_name = process_pubsub_message(data)

    return process_chunker_data(message_data, metadata, vector_name)

def process_chunker_data(message_data, metadata, vector_name):

    if metadata:
        metadata["vector_name"] = vector_name

    if message_data is None:
        log.warning(f"No message_data was found in data: {metadata=}")
        return

    log.debug(f"Found metadata in pubsub: {metadata=}")

    # checks if only a llamaindex chunking/embedder, return early as no other processing needed
    llamacheck = llamaindex_chunker_check(message_data, metadata, vector_name)
    if llamacheck:
        return llamacheck
    
    # if only a discovery engine memory, return early as no other processing needed
    discovery_check = discovery_engine_chunker_check(message_data, metadata, vector_name)
    if discovery_check:
        return discovery_check

    chunks = []

    if message_data.startswith("gs://"):
        chunks, metadata =  handle_gcs_message(message_data, metadata, vector_name)

    elif message_data.startswith("https://drive.google.com") or message_data.startswith("https://docs.google.com"):
        chunks, metadata = handle_google_drive_message(message_data, metadata, vector_name)

    #TODO: support more git service URLs
    elif message_data.startswith("https://github.com"):
        chunks, metadata = handle_github_message(message_data, metadata, vector_name)
        
    elif message_data.startswith("http"):
        chunks, metadata = handle_http_message(message_data, metadata, vector_name)

    else: 
        chunks, metadata = handle_json_content_message(message_data, metadata, vector_name) 
    
    return format_chunk_return(chunks, metadata, vector_name)


def format_chunk_return(chunks, metadata, vector_name):
    # to be really sure
    if metadata:
        metadata["vector_name"] = vector_name

        if metadata.get("return_chunks"):
            log.info("attributes.return_chunks=True detected, skipping process chunks queue")
            output_list = []
            if chunks:
                for chunk in chunks:
                    output_list.append({"page_content": chunk.page_content, "metadata": chunk.metadata})
                    
            return output_list

    process_docs_chunks_vector_name(chunks, vector_name, metadata)

    return metadata


